﻿/* AdsorptionChiller.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics;

namespace Popolo.HVAC.HeatSource
{
  /// <summary>吸着式冷凍機</summary>
  public class AdsorptionChiller
  {

    #region 定数宣言

    /// <summary>定格運転時吸着含水率[kg/kg]</summary>
    private const double NOM_AD0 = 0.04;

    /// <summary>定格運転時脱着含水率[kg/kg]</summary>
    private const double NOM_DS0 = 0.16;

    /// <summary>最大含水率[kg/kg]</summary>
    private const double MAX_W = 0.16;

    /// <summary>最小含水率[kg/kg]</summary>
    private const double MIN_W = 0.04;

    /// <summary>吸着反応器の熱通過有効度[-]</summary>
    private const double EPSILON_DS = 0.9;
    private const double EPSILON_AD = 0.9;

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    /// <summary>熱回収の熱通過率[-]</summary>
    private const double EPSILON_RCV = 0.3;

    /// <summary>蒸発潜熱近似係数</summary>
    private const double CGAM_A = -2.4564;
    private const double CGAM_B = 3.1753e3;

    /// <summary>吸着材温度・含水率・飽和温度の線形近似式係数</summary>
    private readonly double[] ATWS0 = new double[] { -2.6797e-4, 2.2301e-1, -3.0832e1 };
    private readonly double[] ATWS1 = new double[] { 7.6246e-5, 9.3167e-1, -6.3175 };
    
    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>定格周期に対する周期の比率</summary>
    private double ctRate = 1.0;

    /// <summary>蒸発器の伝熱性能[kW/K]</summary>
    private double evaporatorKA;

    /// <summary>凝縮器の伝熱性能[kW/K]</summary>
    private double condensorKA;

    /// <summary>冷水最小流量比[kg/s]</summary>
    private double chilledWaterMinimumFLowRatio = 0.4;

    /// <summary>冷却水最小流量比[kg/s]</summary>
    private double coolingWaterMinimumFLowRatio = 0.4;

    /// <summary>温水最小流量比[kg/s]</summary>
    private double hotWaterMinimumFLowRatio = 0.4;

    /// <summary>熱損失率[-]</summary>
    private double heatLossRate = 0.0;

    /// <summary>サイクル周期に対する吸着材質量[kg/s]</summary>
    private double massPerCycle;

    /// <summary>吸着再生器の熱容量（吸着材質量あたり）[kJ/(kgK)]</summary>
    private double heatCapacity;

    /// <summary>定格の吸着脱着周期[sec]を取得する</summary>
    public double NominalCyclingTime { get; private set; }

    /// <summary>定格周期に対する周期の比率[-]を設定・取得する</summary>
    public double CyclingTimeRate
    {
      get { return ctRate; }
      set { ctRate = Math.Max(0.3, value); }
    }

    /// <summary>冷水出口温度[C]を取得する</summary>
    public double ChilledWaterOutletTemperature { get; private set; }

    /// <summary>冷水出口温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterOutletSetPointTemperature { get; set; }

    /// <summary>冷水入口温度[C]を取得する</summary>
    public double ChilledWaterInletTemperature { get; private set; }

    /// <summary>冷却水出口温度[C]を取得する</summary>
    public double CoolingWaterOutletTemperature { get; private set; }

    /// <summary>冷却水入口温度[C]を取得する</summary>
    public double CoolingWaterInletTemperature { get; private set; }

    /// <summary>温水出口温度[C]を取得する</summary>
    public double HotWaterOutletTemperature { get; private set; }

    /// <summary>温水入口温度[C]を取得する</summary>
    public double HotWaterInletTemperature { get; private set; }

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷却水流量[kg/s]を取得する</summary>
    public double CoolingWaterFlowRate { get; private set; }

    /// <summary>温水流量[kg/s]を取得する</summary>
    public double HotWaterFlowRate { get; private set; }

    /// <summary>冷水定格流量[kg/s]を取得する</summary>
    public double NominalChilledWaterFlowRate { get; private set; }

    /// <summary>冷却水定格流量[kg/s]を取得する</summary>
    public double NominalCoolingWaterFlowRate { get; private set; }

    /// <summary>温水定格流量[kg/s]を取得する</summary>
    public double NominalHotWaterFlowRate { get; private set; }

    /// <summary>定格冷凍能力[kW]を取得する</summary>
    public double NominalCapacity { get; private set; }

    /// <summary>冷水最小流量比[kg/s]を設定・取得する</summary>
    public double ChilledWaterMinimumFLowRatio
    {
      get { return chilledWaterMinimumFLowRatio; }
      private set { chilledWaterMinimumFLowRatio = Math.Min(1, Math.Max(0.4, value)); }
    }

    /// <summary>冷却水最小流量比[kg/s]を設定・取得する</summary>
    public double CoolingWaterMinimumFLowRatio
    {
      get { return coolingWaterMinimumFLowRatio; }
      private set { coolingWaterMinimumFLowRatio = Math.Min(1, Math.Max(0.4, value)); }
    }

    /// <summary>温水最小流量比[kg/s]を設定・取得する</summary>
    public double HotWaterMinimumFLowRatio
    {
      get { return hotWaterMinimumFLowRatio; }
      private set { hotWaterMinimumFLowRatio = Math.Min(1, Math.Max(0.4, value)); }
    }

    /// <summary>冷凍能力[kW]を取得する</summary>
    public double CoolingLoad { get; private set; }

    /// <summary>COP[-]を取得する</summary>
    public double COP
    {
      get
      {
        if (HotWaterFlowRate == 0) return 0;
        else
          return CoolingLoad / (HotWaterFlowRate * WATER_SPECIFIC_HEAT
         * (HotWaterInletTemperature - HotWaterOutletTemperature));
      }
    }

    /// <summary>吸着温度[C]を設定・取得する</summary>
    public double AdsorptionTemperature { get; private set; }

    /// <summary>脱着温度[C]を設定・取得する</summary>
    public double DesorptionTemperature { get; private set; }

    /// <summary>吸着時含水率[kg/kg]を設定・取得する</summary>
    public double WaterContent_Adsorption { get; private set; }

    /// <summary>脱着時含水率[kg/kg]を設定・取得する</summary>
    public double WaterContent_Desorption { get; private set; }

    #endregion

    /// <summary>コンストラクタ</summary>
    /// <param name="chilledWaterInletTemperature">冷水入口温度[C]</param>
    /// <param name="chilledWaterOutletTemperature">冷水出口温度[C]</param>
    /// <param name="chilledWaterFlowRate">冷水質量流量[kg/s]</param>
    /// <param name="coolingWaterInletTemperature">冷却水入口温度[C]</param>
    /// <param name="coolingWaterOutletTemperature">冷却水出口温度[C]</param>
    /// <param name="coolingWaterFlowRate">冷却水質量流量[kg/s]</param>
    /// <param name="hotWaterInletTemperature">温水入口温度[C]</param>
    /// <param name="hotWaterOutletTemperature">温水出口温度[C]</param>
    /// <param name="hotWaterFlowRate">温水質量流量[kg/s]</param>
    public AdsorptionChiller
      (double chilledWaterInletTemperature, double chilledWaterOutletTemperature,
      double chilledWaterFlowRate, double coolingWaterInletTemperature,
      double coolingWaterOutletTemperature, double coolingWaterFlowRate,
      double hotWaterInletTemperature, double hotWaterOutletTemperature,
      double hotWaterFlowRate)
    {
      NominalChilledWaterFlowRate = chilledWaterFlowRate;
      NominalCoolingWaterFlowRate = coolingWaterFlowRate;
      NominalHotWaterFlowRate = hotWaterFlowRate;
      ChilledWaterOutletSetPointTemperature = chilledWaterOutletTemperature;
      CyclingTimeRate = 1.0;

      //絶対温度に変換
      double tHwi = hotWaterInletTemperature + 273.15;
      double tCDwi = coolingWaterInletTemperature + 273.15;
      double tCHwi = chilledWaterInletTemperature + 273.15;
      double tCDwo = coolingWaterOutletTemperature + 273.15;
      double tHwo = hotWaterOutletTemperature + 273.15;
      double tCHwo = chilledWaterOutletTemperature + 273.15;
      double tDS = tHwi - (1- heatLossRate) * (tHwi - tHwo) / EPSILON_DS;

      //熱容量流量の計算
      double mcH = WATER_SPECIFIC_HEAT * hotWaterFlowRate;
      double mcCD = WATER_SPECIFIC_HEAT * coolingWaterFlowRate;
      double mcCH = WATER_SPECIFIC_HEAT * chilledWaterFlowRate;

      //熱収支の確認：熱損失率の計算
      double qCDW = mcCD * (tCDwo - tCDwi);
      double qHW = mcH * (tHwi - tHwo);
      double qEvp = mcCH * (tCHwi - tCHwo);
      double qDS = qCDW - qEvp;
      heatLossRate = 1 - qDS / qHW; //熱収支を調整

      //吸着器・脱着器処理熱量の計算
      double gamTDS = (CGAM_A * tDS + CGAM_B);
      double bb = (EPSILON_AD - 1) * tCDwi - tCDwo + CGAM_B / CGAM_A * EPSILON_AD;
      double cc = (qEvp * EPSILON_AD * gamTDS) / (mcCD * CGAM_A)
        - (EPSILON_AD - 1) * tCDwi * tCDwo - CGAM_B / CGAM_A * EPSILON_AD * tCDwo;
      double tADwo = 0.5 * (-bb - Math.Sqrt(bb * bb - 4 * cc));

      double qAD = (tADwo - tCDwi) * mcCD;
      double tAD = qAD / (mcCD * EPSILON_AD) + tCDwi;
      double qCnd = (tCDwo - tADwo) * mcCD;

      //蒸発器の伝熱係数
      double atws0 = tAD * (tAD * ATWS0[0] + ATWS0[1]) + ATWS0[2];
      double atws1 = tAD * (tAD * ATWS1[0] + ATWS1[1]) + ATWS1[2];
      double awad = mcCH * atws0 / (CGAM_A * tAD + CGAM_B);
      double bwad = (tCHwi - atws1) / atws0;
      double ademt = -Math.Log((NOM_DS0 - bwad) / (NOM_AD0 - bwad));
      double aveWad = bwad - 1d / ademt * (NOM_AD0 - bwad) * (Math.Exp(-ademt) - 1);
      double aveEvp = atws0 * aveWad + atws1;
      double epsEvp = qEvp / (mcCH * (tCHwi - aveEvp));
      evaporatorKA = -Math.Log(1 - epsEvp) * mcCH;

      //凝縮器の熱通過有効度[-]
      atws0 = tDS * (tDS * ATWS0[0] + ATWS0[1]) + ATWS0[2];
      atws1 = tDS * (tDS * ATWS1[0] + ATWS1[1]) + ATWS1[2];
      double awds = mcCD * atws0 / (CGAM_A * tDS + CGAM_B);
      double bwds = (tADwo - atws1) / atws0;
      double dsemt = -Math.Log((NOM_AD0 - bwds) / (NOM_DS0 - bwds));
      double aveWds = bwds - 1d / dsemt * (NOM_DS0 - bwds) * (Math.Exp(-dsemt) - 1);
      double aveCnd = atws0 * aveWds + atws1;
      double epsCnd = qCnd / (mcCD * (aveCnd - tADwo));
      condensorKA = -Math.Log(1 - epsCnd) * mcCD;

      //周期あたりの吸着材質量[kg/sec]
      double mpc1 = awad * epsEvp / ademt;
      double mpc2 = awds * epsCnd / dsemt;
      massPerCycle = Math.Min(mpc1, mpc2);

      //吸着反応器の熱容量
      double qloss = qAD - qEvp;
      heatCapacity = qloss / (massPerCycle * (1 - EPSILON_RCV) * (tDS - tAD));
    }

    /// <summary>状態を更新する</summary>
    /// <param name="chilledWaterInletTemperature">冷水入口温度[C]</param>
    /// <param name="chilledWaterFlowRate">冷水質量流量[kg/s]</param>
    /// <param name="coolingWaterInletTemperature">冷却水入口温度[C]</param>
    /// <param name="coolingWaterFlowRate">冷却水質量流量[kg/s]</param>
    /// <param name="hotWaterInletTemperature">温水入口温度[C]</param>
    /// <param name="hotWaterFlowRate">温水質量流量[kg/s]</param>
    public void Update
      (double chilledWaterInletTemperature, double chilledWaterFlowRate,
      double coolingWaterInletTemperature, double coolingWaterFlowRate,
      double hotWaterInletTemperature, double hotWaterFlowRate)
    {
      if (chilledWaterFlowRate <= 0 || coolingWaterFlowRate <= 0 || hotWaterFlowRate <= 0)
      {
        ShutOff();
        return;
      }

      //状態値を保存
      this.ChilledWaterInletTemperature = chilledWaterInletTemperature;
      this.CoolingWaterInletTemperature = coolingWaterInletTemperature;
      this.HotWaterInletTemperature = hotWaterInletTemperature;
      double rch = chilledWaterFlowRate / NominalChilledWaterFlowRate;
      this.ChilledWaterFlowRate = Math.Max(ChilledWaterMinimumFLowRatio,
        Math.Min(1.4, rch)) * NominalChilledWaterFlowRate;
      double rcd = coolingWaterFlowRate / NominalCoolingWaterFlowRate;
      this.CoolingWaterFlowRate = Math.Max(CoolingWaterMinimumFLowRatio,
        Math.Min(1.4, rcd)) * NominalCoolingWaterFlowRate;
      double rht = hotWaterFlowRate / NominalHotWaterFlowRate;
      this.HotWaterFlowRate = Math.Max(HotWaterMinimumFLowRatio,
        Math.Min(1.4, rht)) * NominalHotWaterFlowRate;

      //冷却運転
      if (ChilledWaterOutletSetPointTemperature < chilledWaterInletTemperature)
      {
        //絶対温度に変換
        double thwi = HotWaterInletTemperature + 273.15;
        double tcwi = CoolingWaterInletTemperature + 273.15;
        double tchwi = ChilledWaterInletTemperature + 273.15;

        //熱容量流量の計算
        double mcH = WATER_SPECIFIC_HEAT * HotWaterFlowRate;
        double mcCD = WATER_SPECIFIC_HEAT * CoolingWaterFlowRate;
        double mcCH = WATER_SPECIFIC_HEAT * ChilledWaterFlowRate;
        double epsEvp = 1 - Math.Exp(-evaporatorKA / mcCH);
        double epsCnd = 1 - Math.Exp(-condensorKA / mcCD);

        //サイクル時間補正
        double mperc = massPerCycle / CyclingTimeRate;

        //吸着温度を収束計算
        double qEVP, tDS, tAD, qAD, qDS, qCND, wad0, wadt, tADwo;
        qEVP = tDS = tAD = qAD = qDS = qCND = wad0 = wadt = tADwo = 0;
        Roots.ErrorFunction eFnc = delegate (double cop)
        {
          //熱源全体の熱収支式から脱着器投入熱を収束計算
          double gamAD, gamDS;
          gamAD = gamDS = 0;
          Roots.ErrorFunction eFnc1 = delegate (double thwo)
          {
            double qHW = mcH * (thwi - thwo);
            qDS = qHW * (1 - heatLossRate);
            qEVP = cop * qHW;
            tDS = thwi - (1 - heatLossRate) * (thwi - thwo) / EPSILON_DS;
            double mtcap = mperc * (1 - EPSILON_RCV) * heatCapacity;
            tAD = (qEVP + EPSILON_AD * mcCD * tcwi + mtcap * tDS) / (EPSILON_AD * mcCD + mtcap);
            gamDS = CGAM_A * tDS + CGAM_B;
            gamAD = CGAM_A * tAD + CGAM_B;
            qCND = qEVP * (gamDS / gamAD);
            qAD = EPSILON_AD * mcCD * (tAD - tcwi);
            return (qCND + qAD) - (qDS + qEVP);
          };
          double tDSwo = Roots.Newton(eFnc1, thwi - 1, 0.000001, 0.000001, 0.00001, 10);
          tADwo = tcwi + EPSILON_AD * (tAD - tcwi);

          //吸着器入口含水率を収束計算
          double atws0ad = tAD * (tAD * ATWS0[0] + ATWS0[1]) + ATWS0[2];
          double atws1ad = tAD * (tAD * ATWS1[0] + ATWS1[1]) + ATWS1[2];
          double awad = mcCH * atws0ad / gamAD;
          double bwad = (tchwi - atws1ad) / atws0ad;
          double atws0ds = tDS * (tDS * ATWS0[0] + ATWS0[1]) + ATWS0[2];
          double atws1ds = tDS * (tDS * ATWS1[0] + ATWS1[1]) + ATWS1[2];
          double awds = mcCD * atws0ds / gamDS;
          double bwds = (tADwo - atws1ds) / atws0ds;
          Roots.ErrorFunction eFnc2 = delegate (double ws)
          {
            wadt = Math.Max(ws, bwad + Math.Exp(-awad * epsEvp / mperc) * (ws - bwad));
            if (MAX_W < wadt) correctAdsorption(ref wadt, tAD, atws0ad, atws1ad);
            wad0 = Math.Min(wadt, (bwds + Math.Exp(-awds * epsCnd / mperc) * (wadt - bwds)));
            if (wad0 < MIN_W) correctAdsorption(ref wad0, tDS, atws0ds, atws1ds);
            return wad0 - ws;
          };
          wad0 = Roots.NewtonBisection(eFnc2, MIN_W, 0.000001, 0.000001, 0.00001, 20);

          return (qEVP / (gamAD * mperc)) - (wadt - wad0);
        };
        Roots.NewtonBisection(eFnc, 0.45, 0.001, 0.001, 0.0001, 20);

        ChilledWaterOutletTemperature = ChilledWaterInletTemperature - qEVP / mcCH;
        HotWaterOutletTemperature = HotWaterInletTemperature - qDS / (1 - heatLossRate) / mcH;
        CoolingWaterOutletTemperature = CoolingWaterInletTemperature + (qCND + qAD) / mcCD;
        CoolingLoad = qEVP;
        WaterContent_Desorption = wad0;
        WaterContent_Adsorption = wadt;
      }
      //運転停止
      else ShutOff();
    }

    private void correctAdsorption
      (ref double adsorption, double temperature, double atws0, double atws1)
    {
      if (MAX_W < adsorption || adsorption < MIN_W)
      {
        double tws = atws0 * adsorption + atws1;
        double rp = Water.GetSaturationPressure(tws - 273.15) / Water.GetSaturationPressure(temperature - 273.15);
        adsorption = getAdsorption(temperature, Math.Min(rp, 1.0));
      }
    }

    /// <summary>運転を停止させる</summary>
    public void ShutOff()
    {
      ChilledWaterOutletTemperature = ChilledWaterInletTemperature;
      CoolingWaterOutletTemperature = CoolingWaterInletTemperature;
      HotWaterOutletTemperature = HotWaterInletTemperature;
      ChilledWaterFlowRate = CoolingWaterFlowRate = HotWaterFlowRate = 0;
      CoolingLoad = 0;
    }

    /// <summary>含水率[kg/kg]を計算する</summary>
    /// <param name="temperature">吸着材温度[K]</param>
    /// <param name="relativePressure">相対圧[kPa]</param>
    /// <returns>含水率[kg/kg]</returns>
    private static double getAdsorption
      (double temperature, double relativePressure)
    {
      double aw0 = temperature * (2.6540e-6 * temperature - 1.9572e-3) + 5.3561e-1;
      double aw1 = temperature * (-4.0912e-3 * temperature + 2.7430) - 5.5037e2;
      double aw2 = temperature * (7.2946e-4 * temperature - 2.9366e-1) + 5.1854e1;
      double aw3 = temperature * (-3.8458e-6 * temperature + 2.1582e-3) - 2.5674e-1;
      return aw0 / (1 + Math.Exp(aw1 * relativePressure + aw2))
        + aw3 * relativePressure;
    }

  }
}
